#include <linux/fs.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/ioport.h>
#include <linux/io.h>

#define LWHPS2FPGA_BRIDGE_BASE  0xff200000
#define EXOR_GPIO_OFS           0x00010000
#define EXOR_GPIO_BASE          (LWHPS2FPGA_BRIDGE_BASE+EXOR_GPIO_OFS)
#define EXOR_GPIO_OUT_IDX       0x00000020
#define EXOR_GPIO_IN_IDX        0x00000010
#define EXOR_GPIO_SIZE          PAGE_SIZE

char *exor_gpio_mem;


static ssize_t exor_gpio_read(struct file * file, char * buf, 
			  size_t count, loff_t *ppos) {
	u32 data;
	char tmp[10];
	ssize_t len;
	
	printk(KERN_DEBUG "EXOR GPIO: Read invoked\n");
	if (*ppos != 0)
	{
		printk(KERN_DEBUG "EXOR GPIO: Invalid position\n");
		return 0;
	}
	data = ioread32(exor_gpio_mem+EXOR_GPIO_IN_IDX);
	
	memset(tmp, 0x00, sizeof(tmp));
	len = snprintf(tmp, sizeof(tmp)-1, "%d", (int)data);
	
	if (count < len)
	{
		printk(KERN_INFO "EXOR GPIO: Not enough space in user buffer: expected %d, got %d\n", (int)len, (int)count);
		return -EINVAL;
	}
	
	if (copy_to_user(buf, tmp, len))
		return -EINVAL;

	*ppos = len;
 
	return len;
}

static ssize_t exor_gpio_write(struct file * file, const char * buf, 
			  size_t count, loff_t *ppos) 
{
	u32 data;
	char* eol;
	char tmp[10];
	
	printk(KERN_DEBUG "EXOR GPIO: Write invoked\n");
	if (buf == NULL) {
		printk(KERN_ERR "EXOR_GPIO: Error, string must not be NULL\n");
		return -EINVAL;
	}
	
	memset(tmp, 0x00, sizeof(tmp));
	if (copy_from_user(tmp, buf, sizeof(tmp)-1))
		return -EINVAL;
	
	eol = strchr(tmp, '\n');
	if (eol)
		*eol = '\0';  
	eol = strchr(tmp, '\r');
	if (eol)
		*eol = '\0';  

	if (kstrtou32(tmp, 10, &data) < 0) {
		printk(KERN_ERR "EXOR GPIO: Could not convert string to integer (%s)\n", tmp);
		return -EINVAL;
	}

	iowrite32(data, exor_gpio_mem+EXOR_GPIO_OUT_IDX);

	return count;
}

static const struct file_operations exor_gpio_fops = {
	.owner		= THIS_MODULE,
	.read		= exor_gpio_read,
	.write		= exor_gpio_write,
}; 
static struct miscdevice exor_gpio_dev = {
	/*
	 * We don't care what minor number we end up with, so tell the
	 * kernel to just pick one.
	 */
	MISC_DYNAMIC_MINOR,
	/*
	 * Name ourselves /dev/exor_gpio.
	 */
	"exor_gpio",
	/*
	 * What functions to call when a program performs file
	 * operations on the device.
	 */
	&exor_gpio_fops
}; 

static int __init exor_gpio_init(void)
{
	int ret;
	struct resource *res;

	printk(KERN_DEBUG "EXOR GPIO: Installing driver..\n");
	
	ret = misc_register(&exor_gpio_dev);
        if (ret < 0)
	{
	        printk(KERN_ERR "EXOR GPIO: driver_register returned %d\n", ret);
		return ret;
	}
	
	/* request input region */
	res = request_mem_region(EXOR_GPIO_BASE, EXOR_GPIO_SIZE, "exorgpio_in");
	if (res == NULL) {
	       printk(KERN_ERR "EXOR GPIO: request_mem_region(1) returned NULL\n");
		misc_deregister(&exor_gpio_dev);
		return -EBUSY;
	}

	exor_gpio_mem = ioremap(EXOR_GPIO_BASE, EXOR_GPIO_SIZE);
	if (exor_gpio_mem == NULL) {
	       printk(KERN_ERR "EXOR GPIO: request_mem_region returned NULL\n");
		misc_deregister(&exor_gpio_dev);
		release_mem_region(EXOR_GPIO_BASE, EXOR_GPIO_SIZE);
		return -EFAULT;
	}
	
	return 0;
}

static void __exit exor_gpio_exit(void)
{
	misc_deregister(&exor_gpio_dev);

	release_mem_region(EXOR_GPIO_BASE, EXOR_GPIO_SIZE);
	iounmap(exor_gpio_mem);
}

module_init(exor_gpio_init);
module_exit(exor_gpio_exit);

